"""Tests for the compute function."""
import datetime
import os
import unittest
import json

import pandas as pd
from pandas.testing import assert_frame_equal

from timeseriesfeatures.compute import compute
from timeseriesfeatures.process import process


class TestCompute(unittest.TestCase):

    def setUp(self):
        self.dir = os.path.dirname(__file__)

    def test_compute(self):
        rows = 500
        df = pd.DataFrame(data={
            "feature1": [float(x) for x in range(rows)],
            "feature2": [float(x + 1) for x in range(rows)],
        }, index=[
            datetime.datetime(2022, 1, 1) + datetime.timedelta(x) for x in range(rows)
        ])
        features = compute(df, 30, 14)
        #with open("expected.json", "w") as handle:
        #    json.dump(features, handle)
        with open(os.path.join(self.dir, "expected.json")) as handle:
            expected_features = json.load(handle)
        self.assertListEqual(features, expected_features)
        df = process(df, features)
        print(df)
        #df.to_parquet("compute.parquet")
        expected_features_df = pd.read_parquet(os.path.join(self.dir, "compute.parquet"))
        assert_frame_equal(df, expected_features_df)



"""Tests for the process function."""
import datetime
import os
import unittest

import pandas as pd
from pandas.testing import assert_frame_equal

from timeseriesfeatures.process import process
from timeseriesfeatures.feature import Feature, FEATURE_TYPE_LAG, FEATURE_TYPE_ROLLING, VALUE_TYPE_NONE, VALUE_TYPE_DAYS
from timeseriesfeatures.transform import Transform


class TestProcess(unittest.TestCase):

    def setUp(self):
        self.dir = os.path.dirname(__file__)

    def test_process(self):
        rows = 100
        df = pd.DataFrame(data={
            "feature1": [float(x) for x in range(rows)],
            "feature2": [float(x + 1) for x in range(rows)],
        }, index=[
            datetime.datetime(2022, 1, 1) + datetime.timedelta(x) for x in range(rows)
        ])
        features = [
            Feature(feature_type=FEATURE_TYPE_LAG, value1=1, transform=str(Transform.NONE), columns=[]),
            Feature(feature_type=FEATURE_TYPE_LAG, value1=2, transform=str(Transform.NONE), columns=[]),
            Feature(feature_type=FEATURE_TYPE_LAG, value1=4, transform=str(Transform.NONE), columns=[]),
            Feature(feature_type=FEATURE_TYPE_LAG, value1=8, transform=str(Transform.NONE), columns=[]),
            Feature(feature_type=FEATURE_TYPE_ROLLING, value1=VALUE_TYPE_NONE, value2=None, transform=str(Transform.NONE), columns=[]),
            Feature(feature_type=FEATURE_TYPE_ROLLING, value1=VALUE_TYPE_DAYS, value2=30, transform=str(Transform.NONE), columns=[]),
        ]
        features_df = process(df, features=features)
        # features_df.to_parquet("expected.parquet")
        expected_features_df = pd.read_parquet(os.path.join(self.dir, "expected.parquet"))
        assert_frame_equal(features_df, expected_features_df)



"""Compute partial auto-correlated features."""

# pylint: disable=duplicate-code
import pandas as pd
from statsmodels.tsa.stattools import pacf  # type: ignore

from .compute_acf import sort_acf_vals
from .feature import FEATURE_TYPE_LAG, VALUE_TYPE_INT, Feature
from .transform import Transform


def compute_pacf(series: pd.Series, max_lag: int, min_lag: int) -> list[Feature]:
    """Compute partial auto-correlated features."""
    acf_vals = pacf(series, nlags=max_lag)
    features = []
    for lag in sort_acf_vals(acf_vals, min_lag):  # type: ignore
        new_feature = Feature(
            feature_type=FEATURE_TYPE_LAG,
            columns=[str(series.name)],
            value1=VALUE_TYPE_INT,
            value2=lag,
            transform=str(Transform.NONE),
        )
        if new_feature in features:
            continue
        features.append(new_feature)
    return features



"""A collection of constants relating to dataframe columns."""

DELIMITER = "_"
LAG_COLUMN = "lag"
TRANSFORM_COLUMN = "transform"
DAYS_COLUMN_SUFFIX = "days"
ALL_SUFFIX = "all"
COUNT_WINDOW_FUNCTION = "count"
SUM_WINDOW_FUNCTION = "sum"
MEAN_WINDOW_FUNCTION = "mean"
MEDIAN_WINDOW_FUNCTION = "median"
VAR_WINDOW_FUNCTION = "var"
STD_WINDOW_FUNCTION = "std"
MIN_WINDOW_FUNCTION = "min"
MAX_WINDOW_FUNCTION = "max"
SKEW_WINDOW_FUNCTION = "skew"
KURT_WINDOW_FUNCTION = "kurt"
SEM_WINDOW_FUNCTION = "sem"
RANK_WINDOW_FUNCTION = "rank"
WINDOW_FUNCTIONS = [
    COUNT_WINDOW_FUNCTION,
    SUM_WINDOW_FUNCTION,
    MEAN_WINDOW_FUNCTION,
    MEDIAN_WINDOW_FUNCTION,
    VAR_WINDOW_FUNCTION,
    STD_WINDOW_FUNCTION,
    MIN_WINDOW_FUNCTION,
    MAX_WINDOW_FUNCTION,
    SKEW_WINDOW_FUNCTION,
    KURT_WINDOW_FUNCTION,
    SEM_WINDOW_FUNCTION,
    RANK_WINDOW_FUNCTION,
]



"""The base class for a feature."""

from typing import NotRequired, TypedDict

FEATURE_TYPE_LAG = "lag"
FEATURE_TYPE_ROLLING = "rolling"
FEATURE_TYPE_NBEATS = "nbeats"

VALUE_TYPE_INT = "int"
VALUE_TYPE_DAYS = "days"
VALUE_TYPE_NONE = "none"


class Feature(TypedDict):
    """A description of a feature to use."""

    feature_type: str
    columns: list[str]
    value1: str | int | None
    value2: NotRequired[str | int | None]
    transform: str
    rank_value: NotRequired[float]
    rank_type: NotRequired[str]



"""Process the NBEATS features."""

# pylint: disable=line-too-long
import datetime
import os
from collections import defaultdict

import pandas as pd
from sktime.forecasting import pytorchforecasting  # type: ignore
from sktime.forecasting.base import ForecastingHorizon  # type: ignore

from .cache import CACHE_FOLDER
from .columns import DELIMITER
from .compute_nbeats import DT_SANITISATION_REPLACEMENTS, NBEATS_MODEL_PREFIX
from .feature import FEATURE_TYPE_NBEATS, Feature


def _desanitise_dt_isoformat(dt: str) -> str:
    for k, v in DT_SANITISATION_REPLACEMENTS.items():
        dt = dt.replace(v, k)
    return dt


def load_model_index(model_dir: str) -> list[tuple[datetime.datetime, str]]:
    """
    Loads the available model timestamps from filenames.
    Assumes files are named like: nbeats_<timestamp>.pkl
    Returns sorted list of (timestamp, path).
    """
    model_files = os.listdir(model_dir)
    models = []
    for file in model_files:
        if not file.endswith(".zip"):
            continue
        name = file.replace(NBEATS_MODEL_PREFIX, "").replace(".zip", "")
        idx = pd.to_datetime(_desanitise_dt_isoformat(name))
        models.append((idx, os.path.join(model_dir, file)))
    return sorted(models, key=lambda x: x[0])


def find_closest_model(
    target_time: datetime.datetime, model_index: list[tuple[datetime.datetime, str]]
):
    """
    Given a list of (timestamp, path), find the latest timestamp <= target_time
    """
    candidates = [m for m in model_index if m[0] <= target_time]
    if not candidates:
        return None
    return candidates[-1][1]  # path of closest earlier model


def _predict_nbeats(series: pd.Series, model_dir: str, min_lag: int) -> pd.Series:
    model_index = load_model_index(model_dir)  # List of (timestamp, path)
    preds = pd.Series(index=series.index, dtype=float)

    # Sort series index so we can batch up by model
    sorted_index = sorted(series.index)
    model_index = sorted(model_index, key=lambda x: x[0])  # sort by model datetime

    # Map each time t to the closest model available before it
    model_to_times = defaultdict(list)

    for t in sorted_index:
        model_path = find_closest_model(t, model_index)
        if model_path is not None:
            model_to_times[model_path].append(t)

    # Predict in batches per model
    for model_path, times in model_to_times.items():
        model = pytorchforecasting.PytorchForecastingNBeats.load_from_path(model_path)

        # This model was trained for a specific forecast time: recover fh from filename or store separately
        # We must re-use the same fh as during training!
        # So: load the .fh file or infer the relative index from training time to prediction time

        # For now, we assume this model was trained with:
        # fh = ForecastingHorizon(range(1, min_lag+1), is_relative=True)
        fh = ForecastingHorizon(range(1, min_lag + 1), is_relative=True)

        pred = model.predict(fh=fh)
        # Fill all requested times with the 1st step-ahead prediction
        for t in times:
            preds.loc[t] = pred.values[0]

    return preds


def process_nbeats(
    df: pd.DataFrame, features: list[Feature], columns: list[str]
) -> pd.DataFrame:
    """Process NBEATS forecasting predictions."""
    if not features:
        return df
    for column in columns:
        for feature in features:
            if feature["feature_type"] != FEATURE_TYPE_NBEATS:
                continue
            if feature["columns"] and column not in feature["columns"]:
                continue
            model_name = feature["value1"]
            if not isinstance(model_name, str):
                continue
            min_lag = feature.get("value2")
            if not isinstance(min_lag, int):
                continue
            df[DELIMITER.join([column, model_name])] = _predict_nbeats(
                df[column], os.path.join(CACHE_FOLDER, model_name), min_lag
            )
    return df



"""Calculate shift features."""

import pandas as pd


def shift_process(df: pd.DataFrame, features: list[str], shift: int) -> pd.DataFrame:
    """Process margins between teams."""
    if shift == 0:
        return df
    df[features] = df[features].shift(shift)
    return df



"""Compute NBEATS features."""

import os

import pandas as pd
from sktime.forecasting import model_selection  # type: ignore
from sktime.forecasting import pytorchforecasting
from sktime.forecasting.base import ForecastingHorizon  # type: ignore

from .cache import CACHE_FOLDER
from .feature import FEATURE_TYPE_NBEATS, Feature
from .transform import Transform

NBEATS_MODEL_EXT = "pkl"
DT_SANITISATION_REPLACEMENTS = {
    ":": "_",
    "/": "-",
}
NBEATS_MODEL_PREFIX = "nbeats_"
NBEATS_FOLDER_PREFIX = "nbeats"


def sanitise_dt_isoformat(column: str) -> str:
    """Sanitise a datetime ISO format."""
    for k, v in DT_SANITISATION_REPLACEMENTS.items():
        column = column.replace(k, v)
    return column


def compute_nbeats(series: pd.Series, max_lag: int, min_lag: int) -> list[Feature]:
    """Compute nbeats features."""
    model_name = "_".join(
        [NBEATS_FOLDER_PREFIX, str(series.name), str(max_lag), str(min_lag)]
    )
    model_folder = os.path.join(CACHE_FOLDER, model_name)
    os.makedirs(model_folder, exist_ok=True)
    splitter = model_selection.SlidingWindowSplitter(
        window_length=365, fh=min_lag, step_length=7, start_with_window=True
    )
    for train_idx, test_idx in splitter.split(series):
        y_train = series.iloc[train_idx]
        target_time = series.index[
            test_idx[0]
        ]  # forecast time index (e.g., Timestamp or int)
        fh = ForecastingHorizon(range(1, min_lag + 1), is_relative=True)
        target_label = sanitise_dt_isoformat(str(target_time))
        model_path = os.path.join(model_folder, f"{NBEATS_MODEL_PREFIX}{target_label}")
        if os.path.exists(model_path + ".zip"):
            continue

        model = pytorchforecasting.PytorchForecastingNBeats(
            trainer_params={"max_epochs": 10, "limit_train_batches": 20},
        )
        if len(y_train) < (max_lag + min_lag):
            continue
        model.fit(y=y_train, fh=fh)
        model.save(model_path, serialization_format="cloudpickle")
    return [
        Feature(
            feature_type=FEATURE_TYPE_NBEATS,
            columns=[str(series.name)],
            value1=model_name,
            value2=min_lag,
            transform=str(Transform.NONE),
        )
    ]



"""The main process function."""

import pandas as pd

from .feature import Feature
from .lag_process import lag_process
from .non_categorical_numeric_columns import \
    find_non_categorical_numeric_columns
from .process_nbeats import process_nbeats
from .rolling_process import rolling_process
from .shift_process import shift_process


def process(
    df: pd.DataFrame,
    features: list[Feature],
    on: str | None = None,
    shift: int = 1,
) -> pd.DataFrame:
    """Process the dataframe for timeseries features."""
    original_columns = df.columns.values.tolist()
    valid_columns = find_non_categorical_numeric_columns(df)
    df = lag_process(df, features, valid_columns)
    lagged_columns = [
        x for x in df.columns.values.tolist() if x not in original_columns
    ]
    df = rolling_process(df, features, on, valid_columns)
    df = process_nbeats(df, features, valid_columns)
    added_features = [
        x
        for x in df.columns.values.tolist()
        if x not in original_columns and x not in lagged_columns
    ]
    df = shift_process(df, valid_columns + added_features, shift)
    return df[sorted(df.columns.values.tolist())]



"""Transform enumeration."""

from enum import StrEnum, auto


class Transform(StrEnum):
    """Transform enumeration."""

    NONE = auto()
    VELOCITY = auto()
    LOG = auto()
    ACCELERATION = auto()
    JERK = auto()
    SNAP = auto()
    SMA_5 = auto()
    CRACKLE = auto()



"""A log transform."""

import numpy as np
import pandas as pd


def log_transform(series: pd.Series) -> pd.Series:
    """Transforms a series by log."""
    return np.log1p(series.clip(lower=0))  # type: ignore



"""Tests for the log transform"""
import os
import unittest

import pandas as pd

from timeseriesfeatures.transform_log import log_transform


class TestLogTransform(unittest.TestCase):

    def setUp(self):
        self.dir = os.path.dirname(__file__)

    def test_log_transform(self):
        series = pd.Series(data=[10.0, 20.0, 30.0])
        transformed_series = log_transform(series)
        self.assertListEqual(transformed_series.to_list(), [2.3978952727983707, 3.044522437723423, 3.4339872044851463])



"""Calculate lag features."""

import pandas as pd

from .columns import DELIMITER, LAG_COLUMN, TRANSFORM_COLUMN
from .feature import FEATURE_TYPE_LAG, Feature
from .transforms import TRANSFORMS


def lag_process(
    df: pd.DataFrame, features: list[Feature], columns: list[str]
) -> pd.DataFrame:
    """Process lags."""
    if not features:
        return df
    for column in columns:
        for feature in features:
            if feature["feature_type"] != FEATURE_TYPE_LAG:
                continue
            if feature["columns"] and column not in feature["columns"]:
                continue
            lag = feature["value1"]
            if not isinstance(lag, int):
                continue
            new_column = DELIMITER.join(
                [column, TRANSFORM_COLUMN, feature["transform"], LAG_COLUMN, str(lag)]
            )
            df[new_column] = TRANSFORMS[feature["transform"]](df[column]).shift(lag)
    return df



"""An derivative transform."""

import functools
from typing import Callable

import pandas as pd


def derivative_transform(series: pd.Series) -> pd.Series:
    """Transforms a series by derivative."""
    return series.pct_change(fill_method=None)


def create_derivative_transform(derivative: int) -> Callable[[pd.Series], pd.Series]:
    """Creates a function determining the Nth order derivative."""

    def _recursive_derivative_transform(
        series: pd.Series, derivative: int
    ) -> pd.Series:
        for _ in range(derivative):
            series = derivative_transform(series)
        return series

    return functools.partial(_recursive_derivative_transform, derivative=derivative)



"""A helper function for retrieving numeric columns without categoricals."""

import numpy as np
import pandas as pd


def find_non_categorical_numeric_columns(df: pd.DataFrame) -> list[str]:
    """
    Finds numeric columns in a Pandas DataFrame that are not categorical.

    Args:
        df (pd.DataFrame): The input DataFrame.

    Returns:
        list: A list of column names that are numeric and not categorical.
    """
    numeric_cols = set(df.select_dtypes(include=np.number).columns.tolist())
    categorical_cols = set(df.select_dtypes(include="category").columns.tolist())
    return sorted(list(numeric_cols.difference(categorical_cols)))



"""Compute auto-correlated features."""

import pandas as pd
from statsmodels.tsa.stattools import acf  # type: ignore

from .feature import FEATURE_TYPE_LAG, VALUE_TYPE_INT, Feature
from .transform import Transform


def sort_acf_vals(acf_vals: list[float], min_lag: int) -> list[int]:
    """Sort auto-correlation values."""
    return [
        lag
        for lag, val in sorted(
            [
                (lag, val)
                for lag, val in enumerate(acf_vals)
                if abs(val) > 0.3 and lag != 0 and lag > min_lag
            ],
            key=lambda x: abs(x[1]),
            reverse=True,
        )[:4]
    ]


def compute_acf(series: pd.Series, max_lag: int, min_lag: int) -> list[Feature]:
    """Compute auto-correlated features."""
    acf_vals = acf(series, nlags=max_lag, fft=False)
    features = []
    for lag in sort_acf_vals(acf_vals, min_lag):  # type: ignore
        new_feature = Feature(
            feature_type=FEATURE_TYPE_LAG,
            columns=[str(series.name)],
            value1=VALUE_TYPE_INT,
            value2=lag,
            transform=str(Transform.NONE),
        )
        if new_feature in features:
            continue
        features.append(new_feature)
    return features



"""A list of transforms."""

from .transform import Transform
from .transform_derivative import create_derivative_transform
from .transform_log import log_transform
from .transform_sma import create_sma_transform

TRANSFORMS = {
    str(Transform.NONE): lambda x: x,
    str(Transform.VELOCITY): create_derivative_transform(1),
    str(Transform.LOG): log_transform,
    str(Transform.ACCELERATION): create_derivative_transform(2),
    str(Transform.JERK): create_derivative_transform(3),
    str(Transform.SNAP): create_derivative_transform(4),
    str(Transform.SMA_5): create_sma_transform(5),
    str(Transform.CRACKLE): create_derivative_transform(5),
}



"""An SMA transform."""

import functools
from typing import Callable

import pandas as pd


def sma_transform(series: pd.Series, window: int) -> pd.Series:
    """Transforms a series by SMA."""
    return series.rolling(window=window).mean()


def create_sma_transform(window: int) -> Callable[[pd.Series], pd.Series]:
    """Creates a function determining the SMA with the window."""
    return functools.partial(sma_transform, window=window)



"""Calculate rolling features."""

# pylint: disable=too-many-branches,too-many-statements
import datetime

import pandas as pd

from .columns import (ALL_SUFFIX, COUNT_WINDOW_FUNCTION, DAYS_COLUMN_SUFFIX,
                      DELIMITER, KURT_WINDOW_FUNCTION, MAX_WINDOW_FUNCTION,
                      MEAN_WINDOW_FUNCTION, MEDIAN_WINDOW_FUNCTION,
                      MIN_WINDOW_FUNCTION, RANK_WINDOW_FUNCTION,
                      SEM_WINDOW_FUNCTION, SKEW_WINDOW_FUNCTION,
                      STD_WINDOW_FUNCTION, SUM_WINDOW_FUNCTION,
                      TRANSFORM_COLUMN, VAR_WINDOW_FUNCTION, WINDOW_FUNCTIONS)
from .feature import (FEATURE_TYPE_ROLLING, VALUE_TYPE_DAYS, VALUE_TYPE_INT,
                      Feature)
from .transforms import TRANSFORMS


def rolling_process(
    df: pd.DataFrame,
    features: list[Feature],
    on: str | None,
    columns: list[str],
) -> pd.DataFrame:
    """Process margins between teams."""
    if not features:
        return df
    for column in columns:
        for feature in features:
            if feature["feature_type"] != FEATURE_TYPE_ROLLING:
                continue
            if feature["columns"] and column not in feature["columns"]:
                continue
            input_type = feature["value1"]
            if "value2" not in feature:
                continue
            input_value = feature["value2"]
            window: int | datetime.timedelta | None = None
            if input_value is not None:
                if input_type == VALUE_TYPE_INT:
                    window = int(input_value)
                elif input_type == VALUE_TYPE_DAYS:
                    window = datetime.timedelta(days=int(input_value))
            window_df = df[[column] + [] if on is None else [on]].copy()
            window_df[column] = TRANSFORMS[feature["transform"]](df[column])
            window_df = (
                window_df.rolling(window, on=on)  # type: ignore
                if window is not None
                else window_df.expanding()
            )
            window_df = window_df[column]  # type: ignore
            window_col = ALL_SUFFIX
            if isinstance(window, int):
                window_col = str(window)
            elif isinstance(window, datetime.timedelta):
                window_col = str(window.days) + DAYS_COLUMN_SUFFIX
            for window_func in WINDOW_FUNCTIONS:
                new_column = DELIMITER.join(
                    [
                        column,
                        TRANSFORM_COLUMN,
                        feature["transform"],
                        window_func,
                        window_col,
                    ]
                )
                if window_func == COUNT_WINDOW_FUNCTION:
                    df[new_column] = window_df.count()
                elif window_func == SUM_WINDOW_FUNCTION:
                    df[new_column] = window_df.sum()
                elif window_func == MEAN_WINDOW_FUNCTION:
                    df[new_column] = window_df.mean()
                elif window_func == MEDIAN_WINDOW_FUNCTION:
                    df[new_column] = window_df.median()
                elif window_func == VAR_WINDOW_FUNCTION:
                    df[new_column] = window_df.var()
                elif window_func == STD_WINDOW_FUNCTION:
                    df[new_column] = window_df.std()
                elif window_func == MIN_WINDOW_FUNCTION:
                    df[new_column] = window_df.min()
                elif window_func == MAX_WINDOW_FUNCTION:
                    df[new_column] = window_df.max()
                elif window_func == SKEW_WINDOW_FUNCTION:
                    df[new_column] = window_df.skew()
                elif window_func == KURT_WINDOW_FUNCTION:
                    df[new_column] = window_df.kurt()
                elif window_func == SEM_WINDOW_FUNCTION:
                    df[new_column] = window_df.sem()
                elif window_func == RANK_WINDOW_FUNCTION:
                    df[new_column] = window_df.rank()
                else:
                    raise ValueError(f"Unrecognised window function: {window_func}")
    return df



"""The function for automatically computing features."""

import pandas as pd

from .compute_acf import compute_acf
from .compute_nbeats import compute_nbeats
from .compute_pacf import compute_pacf
from .feature import Feature
from .non_categorical_numeric_columns import \
    find_non_categorical_numeric_columns


def compute(
    df: pd.DataFrame,
    max_lag: int,
    min_lag: int,
) -> list[Feature]:
    """Process the dataframe for determining timeseries features."""
    features = []
    columns = find_non_categorical_numeric_columns(df)
    for column in columns:
        series = df[column]
        for compute_func in [compute_acf, compute_pacf, compute_nbeats]:
            features.extend(compute_func(series, max_lag, min_lag))
    return features
